
/** @module : Packetizer
 *  @author : Michel Kinsy
 
 *  Copyright (c) 2010 Heracles (CSG/CSAIL/MIT)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */

module packetizer #(parameter CORE = 0, DATA_WIDTH = 32, ADDRESS_BITS = 20, NUM_CACHES = 2, MSG_BITS = 3, OFFSET_BITS= 3,
                    IN_PORTS = 1, OUT_PORTS = 1, VC_BITS = 1, 
                    VC_DEPTH_BITS = 4, ID_BITS = 4, EXTRA  = 2, TYPE_BITS = 2) ( 
					
				     clock, reset, 
				     				     
				     network2mem_msg, 
				     network2mem_address, network2mem_data,
				     
				     mem2network_msg, 
				     mem2network_address, mem2network_data,
				     
				     network2cache_msg, 
				     network2cache_address, network2cache_data,
				     
				     cache2network_msg, 
				     cache2network_address, cache2network_data,
				     
				     from_core_flit, v_from_core, 
				     from_core_empty, from_core_full,
  
  					 to_core_flit, v_to_core,
  					 to_core_empty, to_core_full  
				     
);
localparam VC_PER_PORTS  = (1 << VC_BITS); 
localparam FLOW_BITS     = (2*ID_BITS) + EXTRA;  
localparam FLIT_WIDTH    = FLOW_BITS + TYPE_BITS + VC_BITS + DATA_WIDTH;
localparam OFFSET        = (1<<OFFSET_BITS); 

input clock, reset; 

input [MSG_BITS-1:0]     mem2network_msg ;
input [ADDRESS_BITS-1:0] mem2network_address ;
input [DATA_WIDTH-1:0]   mem2network_data ;

input [(MSG_BITS * NUM_CACHES)-1:0]     cache2network_msg; 
input [(ADDRESS_BITS * NUM_CACHES)-1:0] cache2network_address; 
input [(DATA_WIDTH * NUM_CACHES)-1:0]   cache2network_data ;

output [MSG_BITS-1:0]     network2mem_msg; 
output [ADDRESS_BITS-1:0] network2mem_address; 
output [DATA_WIDTH-1:0]   network2mem_data;

output [(MSG_BITS * NUM_CACHES)-1:0]     network2cache_msg; 
output [(ADDRESS_BITS * NUM_CACHES)-1:0] network2cache_address; 
output [(DATA_WIDTH * NUM_CACHES)-1:0]   network2cache_data;

output [(FLIT_WIDTH * IN_PORTS) -1:0] from_core_flit;
output [IN_PORTS-1: 0] v_from_core;
output [(VC_PER_PORTS * IN_PORTS) -1:0] from_core_empty;
output [(VC_PER_PORTS * IN_PORTS) -1:0] from_core_full; 
  
input  [(FLIT_WIDTH * OUT_PORTS) -1:0] to_core_flit;
input  [OUT_PORTS-1: 0] v_to_core;
input  [(VC_PER_PORTS * IN_PORTS) -1:0] to_core_empty;
input  [(VC_PER_PORTS * IN_PORTS) -1:0] to_core_full;

wire [ADDRESS_BITS-1:0] in_address;
wire [(DATA_WIDTH * OFFSET)-1:0]  in_cache_line;
wire in_valid;
 
wire [ADDRESS_BITS-1:0] out_address;
wire [(DATA_WIDTH * OFFSET)-1:0]  out_cache_line;
wire out_valid;
wire reply2cache;

wire [ADDRESS_BITS-1:0] network_req_address;
wire network_req;

wire read, write, ack;
wire [ADDRESS_BITS-1:0] address;
wire [(DATA_WIDTH * OFFSET)-1:0]  mem_in_cache_line;

wire ready, mem_valid; 
wire [(DATA_WIDTH * OFFSET)-1:0]  mem_out_cache_line;
 
local_mem_interface #(DATA_WIDTH, ADDRESS_BITS, NUM_CACHES, MSG_BITS, OFFSET_BITS) Mem_Interface (
				     clock, reset,  
				   
				     mem2network_msg, 
				     mem2network_address, mem2network_data, 
				     
				     network2mem_msg, 
				     network2mem_address, network2mem_data, 
				     
				     ready, read, write, ack, 
				     address, mem_in_cache_line,
				     mem_valid, mem_out_cache_line
);

local_caches_interface #(DATA_WIDTH, ADDRESS_BITS, NUM_CACHES, MSG_BITS, OFFSET_BITS) Caches_Interface (
				     clock, reset,  
				     				     
				     cache2network_msg, 
				     cache2network_address, cache2network_data,
				     network2cache_msg, 
				     network2cache_address, network2cache_data, 
				     
				     in_address, 
         			 in_cache_line,
         		  	 in_valid,  
            		
         		  	 out_address,
         		  	 out_cache_line,
         		  	 out_valid, 
         		  	 reply2cache,
            		 
         		  	 network_req_address, 
         		   	 network_req
); 

packetizer_lite #(CORE, DATA_WIDTH, ADDRESS_BITS, OFFSET_BITS, IN_PORTS, OUT_PORTS, VC_PER_PORTS, 
                 VC_BITS, VC_DEPTH_BITS, ID_BITS, EXTRA, TYPE_BITS) Real_Packetizer (
				     clock, reset,   		     				     
				     
				     from_core_flit, v_from_core, 
				     from_core_empty, from_core_full,
  
  					 to_core_flit, v_to_core,
  					 to_core_empty, to_core_full,  
				     
				     in_address, 
     		      	 in_cache_line,
     		      	 in_valid,  
            		
     		      	 out_address,
     		      	 out_cache_line,
     		      	 out_valid,
     		      	 reply2cache, 
            		 
     		      	 network_req_address, 
     		      	 network_req,
            		 
				     ready, read, write, ack, 
				     address, mem_in_cache_line,
				     mem_valid, mem_out_cache_line		 
);  

endmodule