
/** @module : Packetizer_helpers
 *  @author : Michel Kinsy
 
 *  Copyright (c) 2010 Heracles (CSG/CSAIL/MIT)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 *
 */
   
module local_mem_interface #(parameter DATA_WIDTH = 32, ADDRESS_BITS = 20, NUM_CACHES = 2, MSG_BITS = 3, OFFSET_BITS= 3) (
				     clock, reset,  
				   
				     mem2network_msg, 
				     mem2network_address, mem2network_data, 
				     
				     network2mem_msg, 
				     network2mem_address, network2mem_data, 
				     
				     ready, read, write, ack, 
				     address, mem_in_cache_line,
				     mem_valid, mem_out_cache_line
);

localparam OFFSET   = 1<<OFFSET_BITS; 

input clock, reset; 

input [MSG_BITS-1:0]      mem2network_msg; 
input [ADDRESS_BITS-1:0]  mem2network_address; 
input [DATA_WIDTH-1:0]    mem2network_data;
 				     
output [MSG_BITS-1:0]     network2mem_msg; 
output [ADDRESS_BITS-1:0] network2mem_address; 
output [DATA_WIDTH-1:0]   network2mem_data; 

output ready;
input  read, write, ack;
input  [ADDRESS_BITS-1:0] address;
input  [(DATA_WIDTH * OFFSET)-1:0]  mem_in_cache_line;

output mem_valid; 
output [(DATA_WIDTH * OFFSET)-1:0]  mem_out_cache_line;

reg [DATA_WIDTH -1:0]  t_mem_out_cache_line [0: (OFFSET -1)];
reg [MSG_BITS-1:0]    t_network2mem_msg; 
reg [ADDRESS_BITS-1:0]t_network2mem_address; 
reg [DATA_WIDTH-1:0]  t_network2mem_data;

reg [OFFSET_BITS-1 :0] zero_offset; 
reg [2:0] state;
reg [OFFSET_BITS :0] counter; 
reg t_ready, t_mem_valid;  
localparam MEM_NO_MSG = 0, MEM_READY = 1, MEM_SENT = 2;
localparam NO_REQ = 0, WB_REQ = 1, C_SENT = 2, C_RECV = 3, R_REQ = 4, W_REQ = 5;
localparam IDLE = 0, WRITE_ST = 1, W_WAIT = 2, READ_ST = 3, R_WAIT = 4, DONE = 5;

assign  network2mem_msg     = t_network2mem_msg; 
assign  network2mem_address = t_network2mem_address; 
assign  network2mem_data    = t_network2mem_data;

genvar i;
generate  
	    for (i = 0; i < OFFSET; i=i+1) begin : MMOCL
      		assign mem_out_cache_line[(((i + 1) *(DATA_WIDTH))-1) -: DATA_WIDTH] = t_mem_out_cache_line[i]; 
  		end   	
endgenerate 
assign  mem_valid           = t_mem_valid; 
assign  ready               = t_ready;
  
integer j;   
always @(posedge clock)
	if(reset==1)begin 
	     t_network2mem_msg     <= NO_REQ;
	     t_network2mem_address <= 0;
	     t_network2mem_data    <= 0;
	     state                 <= IDLE;  
	     counter               <= 0;
		 for (j = 0; j < OFFSET; j=j+1) t_mem_out_cache_line[j]  <= 0;
	     t_mem_valid           <= 0;    
	     t_ready               <= 1;    
	     zero_offset           <= 0; 
	end 
	else  begin      
             case (state)
             	   IDLE: begin
             	         for (j = 0; j < OFFSET; j=j+1) t_mem_out_cache_line[j]        <= 0;
             	         state <= (read == 1)? READ_ST : (write==1)? WRITE_ST : IDLE;
             	         t_ready <= ((read == 1)|| (write==1))? 0 : 1;
                   end
                          
                   READ_ST: begin
                         t_network2mem_msg     <= R_REQ; 
    			         t_network2mem_address <= address;   
    			         counter               <= 0;
    			         state                 <= R_WAIT;
                   end
                   
                   
                   R_WAIT: begin
                         if ((mem2network_msg == MEM_SENT) && (t_network2mem_address == mem2network_address)) begin 
    	                     if(counter < OFFSET) begin
    	                        t_network2mem_msg     <= C_RECV;
    	                        counter               <= counter + 1;
    	                        t_mem_out_cache_line[counter] <= mem2network_data;      
    	                     end
    	                     else begin
    	                        t_network2mem_msg     <= NO_REQ;
    	                     end
                         end
    	                    
    	                 if ((mem2network_msg == MEM_NO_MSG) && (counter == OFFSET))begin
						              state                      <= DONE;
						              t_mem_valid                <= 1; 
						              t_network2mem_msg          <= NO_REQ;
    	                 end
    	                 else state                     <= R_WAIT;
                   end

                    DONE: begin
                          state                            <= ack? IDLE : DONE;
                          t_mem_valid                      <= ack? 0 : 1; 
                   end 
                                
                    WRITE_ST : begin
                         t_network2mem_msg     <= WB_REQ;    
    			         t_network2mem_address <= address;
    			         counter               <= 0;
    			         state                 <= W_WAIT;
                   end 
                     
                  W_WAIT: begin
                         if (mem2network_msg == MEM_READY) begin      
    	                     if(counter < OFFSET) begin
    	                        t_network2mem_msg     <= C_SENT;
    	                        counter               <= counter + 1;
    	                        t_network2mem_address <= address  + counter;
    	                        t_network2mem_data    <= mem_in_cache_line[(((counter + 1)* DATA_WIDTH) -1) -: DATA_WIDTH]; 
    	                        state                 <= W_WAIT;     
    	                    end
    	                    else begin
    	                        t_network2mem_msg     <= NO_REQ;
    	                        t_ready               <= 1; 
    	                        state                 <= IDLE;
    	                    end
    	                 end
    	                 else state      <= W_WAIT;                    
                   end                    
        		endcase
	end
/*
	always @ (posedge clock) begin
    $display ("-----------------------------Local_Mem_Interface---------------------------------"); 
    $display ("State [%d]\t\t| Address [%h]", state, address); 
    $display ("MEM2NET MSG [%d]\t| MEM2NET Add [%h]\t| MEM2NET DATA [%h]", 
              mem2network_msg, mem2network_address, mem2network_data); 
    $display ("NET2MEM MSG [%d]\t| NET2MEM Add [%h]\t| NET2MEM DATA [%h]", 
              network2mem_msg,network2mem_address, network2mem_data);
    $display ("----------------------------------------------------------------------------------"); 
  end
//*/
endmodule


module cache_line #(parameter DATA_WIDTH = 32, ADDRESS_BITS = 20, OFFSET_BITS= 3) (
		clock, address, data_in, rEn, 
		wEn, data_out, valid, 
		
		in_address, 
		in_cache_line,
		in_valid,  
		
		out_address,
		out_cache_line
); 

localparam OFFSET   = 1<<OFFSET_BITS;

input clock, rEn, wEn;
input [ADDRESS_BITS-1:0]  address ;
input [DATA_WIDTH-1:0]  data_in;
output valid; 
output [DATA_WIDTH-1:0] data_out;
 
input [ADDRESS_BITS-1:0] in_address;
input [(DATA_WIDTH * OFFSET)-1:0]  in_cache_line;
input in_valid;
input [ADDRESS_BITS-1:0] out_address;
output [(DATA_WIDTH * OFFSET)-1:0]  out_cache_line;

reg  [DATA_WIDTH-1:0]   from_cache_data [0:OFFSET-1];
wire [DATA_WIDTH-1:0]   to_cache_data   [0:OFFSET-1];
wire [OFFSET_BITS-1:0]  c_select  =  address - in_address;
wire [OFFSET_BITS-1:0]  n_select  =  address - out_address;
 
	//--------------Code Starts Here------------------ 
    always @ (posedge clock) begin : LINE_WRITE
       if (wEn ) begin
          from_cache_data[n_select]  <= data_in;
       end
    end
    
    //----------------------------------------------------
    // Drive the outputs
    //----------------------------------------------------
	assign data_out = (rEn & ~wEn)?  to_cache_data[c_select] : 0; 
	assign valid    = (rEn & ~wEn & in_valid)?  1 : 0;
	
	genvar i;
  	generate   
	    for (i = 0; i < OFFSET; i=i+1) begin : CN_CONNECT
      		assign out_cache_line[(((i + 1) *(DATA_WIDTH))-1) -: DATA_WIDTH]= from_cache_data[i];
      		assign to_cache_data[i]    = in_cache_line[(((i + 1) *(DATA_WIDTH))-1) -: DATA_WIDTH];
  		end 
  	endgenerate
  
endmodule


module local_caches_interface #(parameter DATA_WIDTH = 32, ADDRESS_BITS = 20, NUM_CACHES = 2, MSG_BITS = 3, OFFSET_BITS= 3) (
				     clock, reset,  
				     				     
				     cache2network_msg, 
				     cache2network_address, cache2network_data,
				     network2cache_msg, 
				     network2cache_address, network2cache_data, 
				     
				     in_address, 
         		  	 in_cache_line,
         		  	 in_valid,  
            		
         		 	 out_address,
         		  	 out_cache_line,
         		  	 out_valid, 
         		  	 reply2cache,
            		 
         		  	 network_req_address, 
         		  	 network_req
); 

localparam OFFSET   = 1<<OFFSET_BITS;

input clock, reset; 				     
input [(MSG_BITS * NUM_CACHES)-1:0]     cache2network_msg; 
input [(ADDRESS_BITS * NUM_CACHES)-1:0] cache2network_address; 
input [(DATA_WIDTH * NUM_CACHES)-1:0]   cache2network_data;

output [(MSG_BITS * NUM_CACHES)-1:0]     network2cache_msg; 
output [(ADDRESS_BITS * NUM_CACHES)-1:0] network2cache_address; 
output [(DATA_WIDTH * NUM_CACHES)-1:0]   network2cache_data;

input [ADDRESS_BITS-1:0] in_address;
input [(DATA_WIDTH * OFFSET)-1:0]  in_cache_line;
input in_valid;
 
output [ADDRESS_BITS-1:0] out_address;
output [(DATA_WIDTH * OFFSET)-1:0]  out_cache_line;
output out_valid;
input reply2cache; 

output [ADDRESS_BITS-1:0] network_req_address;
output network_req;

 //Number of bits to represent cache number. 
 //define the log2 function
function integer log2;
	input integer num;
	integer i, result;
	begin
		for (i = 0; 2 ** i < num; i = i + 1)
			result = i + 1;
		log2 = result;
	end
endfunction
localparam ID_BITS = log2(NUM_CACHES);
  
localparam MEM_NO_MSG = 0, MEM_READY = 1, MEM_SENT = 2;
localparam NO_REQ = 0, WB_REQ = 1, C_SENT = 2, C_RECV = 3, R_REQ = 4, W_REQ = 5;

wire [MSG_BITS-1:0]     w_cache2network_msg     [0: (NUM_CACHES-1)]; 
wire [ADDRESS_BITS-1:0] w_cache2network_address [0: (NUM_CACHES-1)]; 
wire [DATA_WIDTH-1:0]   w_cache2network_data    [0: (NUM_CACHES-1)];
	     
reg  [MSG_BITS-1:0]     t_cache2network_msg     [0: (NUM_CACHES-1)]; 
reg  [ADDRESS_BITS-1:0] t_cache2network_address [0: (NUM_CACHES-1)]; 
reg  [DATA_WIDTH-1:0]   t_cache2network_data    [0: (NUM_CACHES-1)];

reg [MSG_BITS-1:0]      t_network2cache_msg      [0: (NUM_CACHES-1)];
reg [ADDRESS_BITS-1:0]  t_network2cache_address  [0: (NUM_CACHES-1)]; 
reg [DATA_WIDTH-1:0]    t_network2cache_data     [0: (NUM_CACHES-1)];

reg    read, write; 
wire   valid; 
reg    [ADDRESS_BITS-1:0] address;
reg    [DATA_WIDTH-1:0]   data_in;
wire   [DATA_WIDTH-1:0]   data_out;

reg [ADDRESS_BITS-1:0] t_out_address;
reg t_out_valid;
reg [ADDRESS_BITS-1:0] t_network_req_address;
reg t_network_req;

cache_line #(DATA_WIDTH, ADDRESS_BITS, OFFSET_BITS) Cache_Packetizer (
		.clock(clock), 
		.address(address), 
		.data_in(data_in), 
		.rEn(read), 
		.wEn(write), 
		.data_out(data_out), 
		.valid(valid), 

		.in_address(in_address), 
		.in_cache_line(in_cache_line),
		.in_valid(in_valid),
		.out_address(t_out_address),  
		.out_cache_line(out_cache_line)		
);

assign out_address = t_out_address; 
assign out_valid = t_out_valid;
assign network_req_address = t_network_req_address; 
assign network_req = t_network_req; 
 
  genvar i;
  generate   
	    for (i = 0; i < NUM_CACHES; i=i+1) begin : C_CONNECTIONS
      		assign network2cache_msg[(((i + 1) *(MSG_BITS))-1) -: MSG_BITS] = t_network2cache_msg[i]; 
      		assign network2cache_address[(((i + 1) *(ADDRESS_BITS))-1) -: ADDRESS_BITS] = t_network2cache_address[i]; 
      		assign network2cache_data[(((i + 1) *(DATA_WIDTH))-1) -: DATA_WIDTH]= t_network2cache_data[i];
      		
      		assign w_cache2network_msg[i]     = cache2network_msg[(((i + 1) *(MSG_BITS))-1) -: MSG_BITS];
      		assign w_cache2network_address[i] = cache2network_address[(((i + 1) *(ADDRESS_BITS))-1) -: ADDRESS_BITS];
      		assign w_cache2network_data[i]    = cache2network_data[(((i + 1) *(DATA_WIDTH))-1) -: DATA_WIDTH];
  		end 
  endgenerate      

reg [ID_BITS-1:0]  serving; 
reg [OFFSET_BITS :0] counter;  
integer index;
 
localparam IDLE = 0, READING = 1, WRITING = 2, CLEAR = 3; 
reg [1:0] state;
reg outstanding; 

wire ready = ((state == IDLE) & (t_network2cache_msg[serving] != MEM_SENT) 
             & (t_network2cache_msg[serving] != MEM_READY) & (~in_valid)); 

always @(posedge clock)
	if(reset==1)begin 
		serving <= 0;
		counter <= 0; 
		read    <= 0; 
		write   <= 0; 
		data_in <= 0;
		address <= 0;
		
		outstanding <= 0; 
		state       <= IDLE;

		t_out_address         <= 0; 
		t_out_valid           <= 0;
		t_network_req_address <= 0;
		t_network_req         <= 0;

		for (index = 0; index < NUM_CACHES; index=index+1) begin
			t_network2cache_msg     [index]   <=  MEM_NO_MSG;
			t_network2cache_data    [index]   <=  0; 
			t_network2cache_address [index]   <=  0; 
			
			t_cache2network_msg     [index]   <=  MEM_NO_MSG;
			t_cache2network_data    [index]   <=  0; 
			t_cache2network_address [index]   <=  0; 
		end 
	end 
    else begin
		// Buffer messages
		for (index = 0; index < NUM_CACHES; index=index+1) begin
			t_cache2network_msg     [index]   <=  (w_cache2network_msg[index] != MEM_NO_MSG)? 
												   w_cache2network_msg[index]: ((~outstanding) & (serving == index))? MEM_NO_MSG : 
												   t_cache2network_msg[index];
												   
			t_cache2network_data    [index]   <=   (w_cache2network_msg[index] != MEM_NO_MSG)?  
													w_cache2network_data[index] : t_cache2network_data[index]; 
			t_cache2network_address [index]   <=   (w_cache2network_msg[index] != MEM_NO_MSG)? 
													w_cache2network_address[index] : t_cache2network_address[index];
		end 
		
		 case (state)
			   IDLE: begin
					if((~outstanding) & ((t_cache2network_msg[serving] == R_REQ) |
					(t_cache2network_msg[serving] == W_REQ) |(t_cache2network_msg[serving] == WB_REQ))) begin
						if(t_cache2network_msg[serving] == WB_REQ) begin 
						   t_network2cache_msg[serving]   <= MEM_READY; 
						   t_out_valid                    <= 0;
						   t_out_address                  <= t_cache2network_address[serving];
						   counter                        <= 0; 
						   state                          <= WRITING;
						end
						else begin
							state                         <= IDLE;
							if (t_network_req == 0) begin
							   t_network_req_address      <= t_cache2network_address[serving];
							   t_network_req              <= 1; 
							end
						end
						outstanding <= 1; 
					end
					else begin 
						if(outstanding) begin 
							if (in_valid && (in_address == t_cache2network_address[serving])) begin
								counter                    <= 0;
								address                    <= t_cache2network_address[serving];
								read                       <= 1;
								t_network_req              <= 0; 
								state                      <= READING;
							end
						end 
						else begin 
							t_network2cache_msg[serving] <=  MEM_NO_MSG; 
							serving                      <= ((serving +1)== NUM_CACHES)? 0 : serving + 1;
							state                        <= IDLE;
						end
					end
			   end
			   
			   CLEAR: begin
				   if (reply2cache == 1) begin
					  t_out_valid   <= 0;
					  outstanding   <= 0; 
					  state         <= IDLE; 
				   end
				   else state       <= CLEAR; // loop
			   end 
											  
			   READING: begin
						if(counter == 0) begin
								address       <= address + 1;
								counter       <= counter + 1;
								t_network2cache_address[serving] <= t_cache2network_address[serving]; 
								t_network2cache_data[serving]    <= data_out;
								t_network2cache_msg[serving]     <= MEM_SENT;   
						end 
						else 
						 if(counter < (OFFSET-1)) begin
							if(w_cache2network_msg[serving]  == C_RECV )begin
								address        <= address + 1;
								counter        <= counter + 1;
								t_network2cache_address[serving] <= t_cache2network_address[serving]; 
								t_network2cache_data[serving]    <= data_out;
								t_network2cache_msg[serving]     <= MEM_SENT;    
							end
							else begin
								t_network2cache_msg[serving]  <= MEM_NO_MSG;
								state                         <= READING; 
							end      
						   end 
						   else begin 
							 if (counter == (OFFSET-1)) begin
								if(w_cache2network_msg[serving]  == C_RECV )begin
									   t_network2cache_address[serving] <= t_cache2network_address[serving]; 
									   t_network2cache_data[serving]    <= data_out;
									   t_network2cache_msg[serving]     <= MEM_SENT;
									   counter                          <= counter + 1;
									   read                             <= 0;
									   outstanding                      <= 0; 
									   state                            <= IDLE;   
								end
								else begin
									t_network2cache_msg[serving]  <= MEM_NO_MSG;
									state                         <= READING; 
								end      
							   end 
					   end
			   end
			   
			   WRITING: begin
						 if(counter < OFFSET) begin
							state       <= WRITING;
							if(w_cache2network_msg[serving]  == C_SENT )begin
								address        <= t_cache2network_address[serving];
								data_in        <= t_cache2network_data[serving]; 
								write          <= 1;
								counter        <= counter + 1;   
							end
							else begin
								 write         <= 0;
							end        
						 end
						 else begin
							write           <= 0;
							t_out_valid     <= 1; 
							if(w_cache2network_msg[serving]  != C_SENT )begin
								state       <= CLEAR;
							end 
							else state   <= WRITING;  
						 end
			   end
			endcase    
    end
   /* 
	always @ (posedge clock) begin
    $display ("-----------------------------Local_Cache_Interface---------------------------------");
	
	$display ("In_valid [%b]", in_valid); 
	$display ("In_address [%h] ", in_address);
	$display ("In_Cache_line [%h] ", in_cache_line);
				
	$display ("Serving [%d]\t| t_CACHE2NET MSG [%d]", serving, t_cache2network_msg[serving]); 
    $display ("State [%d]\t\t| Req Address [%h]\t| Out Address [%h]", state, network_req_address, out_address); 
    $display ("CACHE2NET MSG [%d]\t| CACHE2NET Add [%h]\t| CACHE2NET DATA [%h]", 
              cache2network_msg, cache2network_address, cache2network_data);     
	$display ("0CACHE2NET MSG [%d]\t| tCACHE2NET Add [%h]\t| tCACHE2NET DATA [%h]", 
              t_cache2network_msg[0], t_cache2network_address[0], t_cache2network_data[0]); 
	$display ("1CACHE2NET MSG [%d]\t| tCACHE2NET Add [%h]\t| tCACHE2NET DATA [%h]", 
              t_cache2network_msg[1], t_cache2network_address[1], t_cache2network_data[1]);
    $display ("NET2CACHE MSG [%d]\t| NET2CACHE Add [%h]\t| NET2CACHE DATA [%h]", 
              network2cache_msg,network2cache_address, network2cache_data);
    $display ("----------------------------------------------------------------------------------"); 
  end
//*/

endmodule
