/** @module : Dir_network interface
 *  @author : Michel Kinsy
 
 *  Copyright (c) 2013 Heracles (CSG/CSAIL/MIT)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
					
module network_interface #(parameter CORE = 0, DATA_WIDTH = 32, ADDRESS_BITS = 20, MSG_BITS = 4, 
					OFFSET_BITS= 3, IN_PORTS = 1, OUT_PORTS = 1, VC_BITS = 1, 
                    VC_DEPTH_BITS = 4, CACHE_ID_BITS = 5, ID_BITS = 4, EXTRA  = 2, TYPE_BITS = 2) (
				     clock, reset,  	     				     
				     
				     from_core_flit, v_from_core, 
				     from_core_empty, from_core_full,
  
  					 to_core_flit, v_to_core,
  					 to_core_empty, to_core_full, 
					 
					 me_ready,
					 me_in_msg,
				     me_in_address, 
         		  	 me_in_cache_line,
            		  
					 me_out_msg,
         		 	 me_out_address,
         		  	 me_out_cache_line,
					 
					 ce_in_msg,
				     ce_in_address, 
         		  	 ce_in_cache_line,
            		
					 ce_out_msg,
         		 	 ce_out_address,
         		  	 ce_out_cache_line	 
);

localparam VC_PER_PORTS  = (1 << VC_BITS); 
localparam FLOW_BITS     = (2*ID_BITS) + EXTRA;  
localparam PAY_WIDTH     = MSG_BITS + DATA_WIDTH;
localparam FLIT_WIDTH    = FLOW_BITS + TYPE_BITS + VC_BITS + PAY_WIDTH;
localparam OFFSETS       = 1<<OFFSET_BITS; 
localparam LINE_BITS     = CACHE_ID_BITS + (DATA_WIDTH * OFFSETS); 
localparam INFO_WIDTH    = MSG_BITS + ADDRESS_BITS + LINE_BITS;

input clock, reset; 

output [(FLIT_WIDTH * IN_PORTS) -1:0] from_core_flit;
output [IN_PORTS-1: 0] v_from_core;
output [(VC_PER_PORTS * IN_PORTS) -1:0] from_core_empty;
output [(VC_PER_PORTS * IN_PORTS) -1:0] from_core_full; 
  
input  [(FLIT_WIDTH * OUT_PORTS) -1:0] to_core_flit;
input  [OUT_PORTS-1: 0] v_to_core;
input  [(VC_PER_PORTS * IN_PORTS) -1:0] to_core_empty;
input  [(VC_PER_PORTS * IN_PORTS) -1:0] to_core_full;

input                    me_ready; 
input [MSG_BITS-1:0]     me_in_msg; 
input [ADDRESS_BITS-1:0] me_in_address;
input [LINE_BITS-1:0]    me_in_cache_line;

output [MSG_BITS-1:0]     me_out_msg; 
output [ADDRESS_BITS-1:0] me_out_address;
output [LINE_BITS-1:0]    me_out_cache_line;

input [MSG_BITS-1:0]     ce_in_msg; 
input [ADDRESS_BITS-1:0] ce_in_address;
input [LINE_BITS-1:0]    ce_in_cache_line;

output [MSG_BITS-1:0]     ce_out_msg; 
output [ADDRESS_BITS-1:0] ce_out_address;
output [LINE_BITS-1:0]    ce_out_cache_line;

reg [MSG_BITS-1:0]     me_out_msg = 0; 
reg [ADDRESS_BITS-1:0] me_out_address = 0; 
reg [LINE_BITS-1:0]    me_out_cache_line = 0; 

reg [MSG_BITS-1:0]     ce_out_msg;  
reg [ADDRESS_BITS-1:0] ce_out_address; 
reg [LINE_BITS-1:0]    ce_out_cache_line; 

localparam NO_MSG = 0;
localparam M_REQ = 1, INV_REQ = 2, M_WB_REQ = 3, SHARE_REQ = 4, WB_INV_REQ = 5; //memory requests
localparam M_READY = 6, MEM_SENT = 7,  W_RESP = 8, EX_RESP = 9, SH_RESP = 10;   //memory responses 
localparam R_REQ = 1, W_REQ = 2, C_WB_REQ = 3;                                  //cache requests
localparam C_READY = 4, C_SENT = 5, C_RECV = 6, INV_RESP = 7, WB_RESP = 8, WB_INV_RESP = 9;
//cache responses 
localparam HEAD = 'b10, BODY = 'b00, TAIL = 'b01, ALL= 'b11;
localparam IDLE = 0, SEARCH_VC = 1, READY = 2, GET_MSG = 3, REFILL = 4, SEND = 5;
localparam CHECK_VC = 1, READ_DELAY = 2, GET_FLIT = 3, PROCESS = 4, READ = 5, DONE = 6;
			  
localparam M2I = 0, M2D = 1, I2M = 2, D2M = 3;  // subflows 

reg [2: 0] r_state;
reg [2: 0] s_state;
reg [OFFSETS-1:0]  r_counter;
reg [OFFSETS-1:0]  s_counter;

reg [FLIT_WIDTH -1:0] t_from_core_flit [0 : IN_PORTS-1];
reg [IN_PORTS-1: 0] t_v_from_core;
wire [VC_PER_PORTS -1:0] t_from_core_empty [0 : IN_PORTS-1];
wire [VC_PER_PORTS -1:0] t_from_core_full [0 : IN_PORTS-1]; 
  
wire  [FLIT_WIDTH -1:0] t_to_core_flit [0 : OUT_PORTS-1];
wire  [OUT_PORTS-1: 0] t_v_to_core;
wire  [VC_PER_PORTS -1:0] t_to_core_empty [0 : OUT_PORTS-1];
wire  [VC_PER_PORTS -1:0] t_to_core_full [0 : OUT_PORTS-1];

reg [FLIT_WIDTH - 1:0] in_flit [0:IN_PORTS-1];
reg [VC_BITS - 1:0] in_vc [0:IN_PORTS-1]; 
reg WEn [0:IN_PORTS-1];
   
reg [VC_BITS - 1:0] read_vc [0:IN_PORTS-1];
reg REn [0:IN_PORTS-1];
  
wire [FLIT_WIDTH - 1:0] out_flit [0:IN_PORTS-1];
wire [VC_BITS - 1:0] out_vc [0:IN_PORTS-1];
wire valid [0:IN_PORTS-1];
  
wire [VC_PER_PORTS -1:0] empty [0:IN_PORTS-1];
wire [VC_PER_PORTS - 1:0] full [0:IN_PORTS-1]; 

reg  ce_req_read, ce_resp_read;
wire ce_req_valid;
wire [INFO_WIDTH-1: 0] ce_req_data;
wire ce_req_full, ce_req_empty;
  
wire ce_resp_valid;
wire [INFO_WIDTH-1: 0] ce_resp_data;
wire ce_resp_full, ce_resp_empty;

wire ce_req_write    =((ce_in_msg == M_REQ) | (ce_in_msg == INV_REQ) | 
					  (ce_in_msg == M_WB_REQ) | (ce_in_msg == SHARE_REQ) | 
					  (ce_in_msg == WB_INV_REQ));
wire [INFO_WIDTH-1: 0] ce_req_in_data = {ce_in_msg, ce_in_address, ce_in_cache_line};
wire ce_resp_write   = ((ce_in_msg == W_RESP) | (ce_in_msg == EX_RESP) | (ce_in_msg == SH_RESP)); 
wire [INFO_WIDTH-1: 0] ce_resp_in_data = {ce_in_msg, ce_in_address, ce_in_cache_line};

reg  me_req_read, me_resp_read;
wire me_req_valid;
wire [INFO_WIDTH-1: 0] me_req_data;
wire me_req_full, me_req_empty;
  
wire me_resp_valid;
wire [INFO_WIDTH-1: 0] me_resp_data;
wire me_resp_full, me_resp_empty;

wire me_req_write    = ((me_in_msg == R_REQ) | (me_in_msg == W_REQ) | 
					   (me_in_msg == C_WB_REQ));
wire [INFO_WIDTH-1: 0] me_req_in_data = {me_in_msg, me_in_address, me_in_cache_line};
wire me_resp_write   = ((me_in_msg == INV_RESP) | (me_in_msg == WB_RESP) | 
                       (me_in_msg == WB_INV_RESP) | (me_in_msg == SH_RESP)); 
wire [INFO_WIDTH-1: 0]  me_resp_in_data = {me_in_msg, me_in_address, me_in_cache_line};

reg [MSG_BITS-1:0]       local_msg; 
reg [ADDRESS_BITS-1:0]   local_address; 
reg [LINE_BITS-1:0]      local_data; 
reg [EXTRA - 1: 0]       local_src; 

wire new_msg  = ce_resp_valid | ce_req_valid | me_resp_valid | me_req_valid; 
wire request  = (local_msg == M_REQ) | (local_msg == INV_REQ) | (local_msg == M_WB_REQ) |
				(local_msg == SHARE_REQ) | (local_msg == WB_INV_REQ) | (local_msg == R_REQ) |
				(local_msg == W_REQ) | (local_msg == C_WB_REQ) | (local_msg == INV_RESP);	
wire response = (local_msg == W_RESP) | (local_msg == EX_RESP) | (local_msg == SH_RESP) | 
				(local_msg == WB_RESP) | (local_msg == WB_INV_RESP); 
wire [(DATA_WIDTH - ADDRESS_BITS) - 1: 0] pad = 0; 
wire [(DATA_WIDTH * OFFSETS)-1:0]  padding = 0;

wire[ID_BITS - 1: 0]   source = CORE; 				
reg [ID_BITS - 1: 0]   s_destination; 
reg [EXTRA - 1: 0]     s_sub_flow;  
reg [TYPE_BITS - 1:0]  s_flit_type;
reg [VC_BITS - 1:0]    s_flit_vc;
reg [PAY_WIDTH - 1:0]  s_flit_payload; 
reg [LINE_BITS-1:0]    s_local_data;
reg [MSG_BITS-1:0]     s_no_msg = 0; 

reg  found; 
reg [VC_BITS - 1:0]    free_vc; 
reg [IN_PORTS-1 :0]    free_port;
integer u, v; 

reg [ID_BITS - 1: 0]   r_source; 
reg [EXTRA - 1: 0]     r_sub_flow; 
reg [FLIT_WIDTH - 1:0] r_flit;
reg [FLOW_BITS - 1:0]  r_flow_ID; 
reg [TYPE_BITS - 1:0]  r_flit_type;
reg [VC_BITS - 1:0]    r_flit_vc;
reg [MSG_BITS - 1:0]   r_flit_pay_msg;
reg [DATA_WIDTH - 1:0] r_flit_pay_data;

reg [VC_BITS - 1:0]    vc; 
reg [IN_PORTS-1 :0]    port;

reg [EXTRA - 1: 0]     r_direction; 
reg [INFO_WIDTH-1:0]   r_local_data;

wire  ce_direction = ((r_state == DONE) & ((r_direction == I2M) | (r_direction == D2M))); 
wire  me_direction = ((r_state == DONE) & ((r_direction == M2I) | (r_direction == M2D))); 

genvar i;
generate
	for (i=0; i < IN_PORTS; i=i+1) begin : IN_CORE_PORTS
		buffer_port #(FLIT_WIDTH, VC_BITS, VC_DEPTH_BITS) INP (
		  clock, reset, 
		  WEn[i], in_flit[i], in_vc [i], 
		  REn[i], 1'b0, read_vc[i],  
		  out_flit[i], out_vc[i], valid[i], 
		  empty[i], full[i]
		);
	end

	for (i = 0; i < OUT_PORTS; i = i+1) begin : OUT_CPORTS
		assign t_to_core_flit[i]    = to_core_flit[((FLIT_WIDTH *(i + 1))-1) -: FLIT_WIDTH];
		assign t_to_core_empty[i]   = to_core_empty[((VC_PER_PORTS *(i + 1))-1) -: VC_PER_PORTS]; 
		assign t_to_core_full[i]    = to_core_full[((VC_PER_PORTS *(i + 1))-1) -: VC_PER_PORTS]; 
	end  
	assign t_v_to_core = v_to_core; 
	
	for (i = 0; i < IN_PORTS; i = i+1) begin : IN_CPORTS
		   assign from_core_flit [(((i + 1) *(FLIT_WIDTH))-1) -: FLIT_WIDTH]  = t_from_core_flit[i];
		   assign from_core_empty[(((i + 1) *(VC_PER_PORTS))-1) -: VC_PER_PORTS] = t_from_core_empty[i];
	 assign from_core_full [(((i + 1) *(VC_PER_PORTS))-1) -: VC_PER_PORTS] = t_from_core_full[i];
	end 	
	assign v_from_core = t_v_from_core;
	
	for (i = 0; i < IN_PORTS; i = i+1) begin : EF_CPORTS
			assign t_from_core_empty[i] = empty[i]; 
			assign t_from_core_full[i]  = full[i]; 
	end   	
endgenerate 
  
fifo  #(INFO_WIDTH, VC_DEPTH_BITS, 0) ce_req_Bufs (
	.clk(clock),
	.reset(reset),
	.write_data(ce_req_in_data),
	.wrtEn(ce_req_write),
	.rdEn(ce_req_read),
	.peek(1'b0),

	.read_data(ce_req_data), 
	.valid(ce_req_valid),			
	.full(ce_req_full), 
	.empty(ce_req_empty) 
);

fifo  #(INFO_WIDTH, VC_DEPTH_BITS, 0) ce_resp_Bufs (
	.clk(clock),
	.reset(reset),
	.write_data(ce_resp_in_data),
	.wrtEn(ce_resp_write),
	.rdEn(ce_resp_read),
	.peek(1'b0),

	.read_data(ce_resp_data), 
	.valid(ce_resp_valid),			
	.full(ce_resp_full), 
	.empty(ce_resp_empty) 
);

fifo  #(INFO_WIDTH, VC_DEPTH_BITS, 0) me_req_Bufs (
	.clk(clock),
	.reset(reset),
	.write_data(me_req_in_data),
	.wrtEn(me_req_write),
	.rdEn(me_req_read),
	.peek(1'b0),

	.read_data(me_req_data), 
	.valid(me_req_valid),			
	.full(me_req_full), 
	.empty(me_req_empty) 
);

fifo  #(INFO_WIDTH, VC_DEPTH_BITS, 0) me_resp_Bufs (
	.clk(clock),
	.reset(reset),
	.write_data(me_resp_in_data),
	.wrtEn(me_resp_write),
	.rdEn(me_resp_read),
	.peek(1'b0),

	.read_data(me_resp_data), 
	.valid(me_resp_valid),			
	.full(me_resp_full), 
	.empty(me_resp_empty) 
);
					
always @(posedge clock) begin 
	if(reset==1)begin 
		local_msg     <= 0;
		local_address <= 0;
		local_data    <= 0;
		local_src     <= 0; 
		
		me_req_read     <= 0; 
		me_resp_read    <= 0;
		
		ce_req_read     <= 0; 
		ce_resp_read    <= 0;
	end 
    else begin
		ce_resp_read   <= ((s_state == READY) & (~ce_resp_empty) & (~ce_resp_read))? 1: 0;  
		ce_req_read    <= ((s_state == READY) & ce_resp_empty & (~ce_resp_read) & 
		                  (~ce_req_empty) & (~ce_req_read))? 1: 0; 

		me_resp_read   <= ((s_state == READY) & ce_resp_empty & ce_req_empty & (~me_resp_empty) 
		                  & (~me_resp_read))? 1: 0; 
		me_req_read    <= ((s_state == READY) & ce_resp_empty & ce_req_empty & me_resp_empty & 
		                  (~me_resp_read) & (~me_req_empty) & (~me_req_read))? 1: 0; 
		
		local_msg     <= ce_resp_valid? ce_resp_data[(INFO_WIDTH-1) -: MSG_BITS]:
						 ce_req_valid?  ce_req_data[(INFO_WIDTH-1) -: MSG_BITS]: 
						 me_resp_valid? me_resp_data[(INFO_WIDTH-1) -: MSG_BITS]:
						 me_req_valid?  me_req_data[(INFO_WIDTH-1) -: MSG_BITS]: NO_MSG; 
						 
		local_address <= ce_resp_valid? ce_resp_data[(INFO_WIDTH - MSG_BITS-1) -: ADDRESS_BITS]:
						 ce_req_valid?  ce_req_data[(INFO_WIDTH - MSG_BITS-1) -: ADDRESS_BITS]: 
						 me_resp_valid? me_resp_data[(INFO_WIDTH - MSG_BITS-1) -: ADDRESS_BITS]:
						 me_req_valid?  me_req_data[(INFO_WIDTH - MSG_BITS-1) -: ADDRESS_BITS]: 0; 
						 
		local_data    <= ce_resp_valid? ce_resp_data[(INFO_WIDTH - MSG_BITS - ADDRESS_BITS-1) : 0]:
						 ce_req_valid?  ce_req_data[(INFO_WIDTH - MSG_BITS - ADDRESS_BITS-1) : 0]: 
						 me_resp_valid? me_resp_data[(INFO_WIDTH - MSG_BITS - ADDRESS_BITS-1) : 0]:
						 me_req_valid?  me_req_data[(INFO_WIDTH - MSG_BITS - ADDRESS_BITS-1) : 0]: 0; 
		local_src     <= ce_resp_valid? 0: ce_req_valid?  1: me_resp_valid? 2: me_req_valid? 3: 0; 
	end 
end

//Message sending logic from local core to network 
always @ (posedge clock) begin  
	if(reset == 1) begin 
		s_destination    <= 0;
		s_sub_flow       <= 0;
		s_flit_type      <= ALL;
		s_flit_vc        <= 0;
		s_flit_payload   <= 0;
		
		found            <= 0; 
      	free_vc          <= 0; 
      	free_port        <= 0; 
		
		s_local_data     <= 0; 
		s_counter        <= 0; 
		s_state          <= IDLE; 
	end 
	else begin 
		case (s_state)
		IDLE: begin
      	 	s_destination    <= 0;
      	 	s_sub_flow       <= 0;
         	s_flit_type      <= ALL;
         	s_flit_vc        <= 0;
         	s_flit_payload   <= 0;
			
			found            <= 0; 
			free_vc          <= 0; 
			free_port        <= 0; 
			s_counter        <= 0; 
			s_local_data     <= 0;
			for(u = 0; u < IN_PORTS; u = u+1) begin
				t_from_core_flit[u]  <= 0;
				t_v_from_core[u]     <= 0; 
			end 
			s_state          <= SEARCH_VC; 
		end
	    SEARCH_VC: begin
            for(u = 0; u < IN_PORTS; u = u+1) begin
              if((t_to_core_empty[u] > 0) & ~found) begin
				for(v = 0; v < VC_PER_PORTS; v = v+1) begin
					if(t_to_core_empty[u][v]) begin 
              			free_vc     <= v;
              			free_port   <= u;
              			found       <= 1; 
            		end 
				end
				end
            end
            s_state     <= (found)? READY : SEARCH_VC;
      	end
		READY: begin
			s_state          <= new_msg? GET_MSG : READY; 
		end
		GET_MSG: begin 	
      	 	s_destination    <= (local_src > 1)? local_address[(ADDRESS_BITS-1) -: ID_BITS] : 
								local_data[(LINE_BITS -1) -: ID_BITS];
      	 	s_sub_flow       <= (local_src > 1)?
								((local_data[(LINE_BITS - ID_BITS -1)])? D2M : I2M) :
								((local_data[(LINE_BITS - ID_BITS -1)])? M2D : M2I);
         	s_flit_type      <= request? ALL : HEAD;
         	s_flit_vc        <= free_vc;
         	s_flit_payload   <= {local_msg, pad, local_address};
			
			s_counter        <= 0;
			s_local_data     <= local_data;			
			s_state          <= SEND; 
		end	
		REFILL: begin 	
			t_from_core_flit[free_port]  <= 0;
			t_v_from_core[free_port]     <= 0; 
			
         	s_flit_type                  <= (s_counter == (OFFSETS -1))? TAIL : BODY;
         	s_flit_payload               <= {s_no_msg, pad, 
								            s_local_data[(((s_counter + 1) *(DATA_WIDTH))-1) -: DATA_WIDTH]};
			s_counter        		     <= s_counter + 1;
			s_state                      <= SEND; 	
		end	
		
		SEND: begin
			if(t_to_core_full[free_port][free_vc]) begin 
				t_from_core_flit[free_port]  <= 0;
				t_v_from_core[free_port]     <= 0; 
				s_state                      <= SEND;
			end 
			else begin 
				t_from_core_flit[free_port]  <= {source, s_destination, s_sub_flow, 
												s_flit_type, s_flit_vc, s_flit_payload};
				t_v_from_core[free_port]     <= 1; 
				s_state                      <= ((s_flit_type == ALL) | (s_flit_type == TAIL))? 
				                                IDLE : REFILL;
			end
		end 
							
		default: s_state <= IDLE;
		endcase 
	end 
end 

//Message recieving logic from local core to network 
always @ (posedge clock) begin
      if (reset) begin 
         	// Resentting 
            for(u= 0; u < IN_PORTS; u = u+1) begin
              	 WEn[u]           = 0;       	 
            end
      end
      else begin    
         	// read in input ports 
            for(u= 0; u < IN_PORTS; u = u+1) begin
            	 WEn[u]         = (t_v_to_core[u]) ? 1 : 0; 
              	 if(t_v_to_core[u]) begin
              		  in_flit[u]     = t_to_core_flit[u];
              		  in_vc [u]      = t_to_core_flit[u][((VC_BITS + PAY_WIDTH) -1) -: VC_BITS]; 
              	 end
            end 
      end
end 

 always @ (posedge clock) begin
    if(reset == 1) begin 
		r_flit			<= 0; 
		r_source        <= 0;  
		r_sub_flow      <= 0;  
		r_flit_type     <= 0; 
		r_flit_vc       <= 0; 
		r_flit_pay_msg  <= NO_MSG; 
		r_flit_pay_data <= 0;

		vc           <= 0; 
		port         <= 0; 
	    r_local_data <= 0; 
		r_direction  <= 0; 
		r_counter    <= 0;
		for(u= 0; u < IN_PORTS; u = u+1) begin
			REn[u]   <= 0;       	 
		end
		r_state      <= 	IDLE; 
    end 
    else begin
      case (r_state)
      	   IDLE: begin
				r_flit			<= 0; 
				r_source        <= 0;  
				r_sub_flow      <= 0;  
				r_flit_type     <= 0; 
				r_flit_vc       <= 0; 
				r_flit_pay_msg  <= NO_MSG; 
				r_flit_pay_data <= 0;				

				vc              <= (vc == (VC_PER_PORTS-1))? 0 : vc + 1;
				port            <= (vc == (VC_PER_PORTS-1))? ((port == (IN_PORTS) -1)? 0 : (port +1)) : port;
 
				r_local_data    <= 0; 
				r_direction     <= 0; 
				r_counter       <= 0;	
				r_state         <= (me_ready & (r_counter == 0))? CHECK_VC : IDLE; 
      	   end

      	   CHECK_VC: begin	     
        	     if(empty[port][vc]== 0) begin
        	         REn[port]      <= 1;  
				     read_vc[port]  <= vc; 
				     r_state        <= READ_DELAY; 
        	     end   
      	       else r_state         <= IDLE;
      	   end 
      	   
			READ: begin
        	     if(empty[port][vc]== 0) begin
        	         REn[port]      <= 1;  
				     r_state        <= READ_DELAY; 
        	     end   
      	       else r_state         <= READ;
			end
			
      	   READ_DELAY: begin	     
				r_state             <= GET_FLIT; 
				REn[port]           <= 0;
      	   end 
      	   
          GET_FLIT: begin
              if(valid[port]==0) 
               	  $display("ERROR!!! READ ERROR MEM PROBLEM!!!!"); 
              else begin
				r_flit			<= out_flit[port]; 
				r_source        <= out_flit[port][(FLIT_WIDTH-1) -: ID_BITS];  
				r_sub_flow      <= out_flit[port][(PAY_WIDTH + VC_BITS + TYPE_BITS + EXTRA-1) -: EXTRA];  
				r_flit_type     <= out_flit[port][(PAY_WIDTH + VC_BITS + TYPE_BITS-1) -: TYPE_BITS]; 
				r_flit_vc       <= out_flit[port][(PAY_WIDTH + VC_BITS -1) -: VC_BITS]; 
				r_flit_pay_msg  <= out_flit[port][(PAY_WIDTH -1) -: MSG_BITS]; 
				r_flit_pay_data <= out_flit[port];  
              end
              r_state           <=  PROCESS;
          end

          PROCESS: begin
			if((r_flit_type == ALL) | (r_flit_type == HEAD)) begin 
				r_local_data  <= ((r_sub_flow == I2M) | (r_sub_flow == D2M))? 
								 {r_flit_pay_msg, r_flit_pay_data[(ADDRESS_BITS-1) : 0], r_source, r_sub_flow[0], padding} : 
								 {r_flit_pay_msg, r_flit_pay_data[(ADDRESS_BITS-1) : 0], source, r_sub_flow[0], padding};
				r_direction   <= r_sub_flow; 
				r_counter     <= 0;
				r_state       <= (r_flit_type == ALL)? DONE : READ;    
			end 
			else begin 
				r_local_data[(((r_counter + 1) *(DATA_WIDTH))-1) -: DATA_WIDTH] <= r_flit_pay_data; 
				r_counter    <= r_counter + 1;
				r_state      <= (r_flit_type == TAIL)? DONE : READ; 
			end 
		 end
		 
		DONE: begin
			r_local_data  <=  0; 
			r_state       <= IDLE;
		end
		 
      	default:    r_state   <= IDLE;
      endcase 
   end 
end

always @ (posedge clock) begin
	ce_out_msg        <= ce_direction? r_local_data[(INFO_WIDTH-1) -: MSG_BITS] : 0; 
	ce_out_address    <= ce_direction? r_local_data[(INFO_WIDTH-MSG_BITS-1) -: ADDRESS_BITS] : 0;
	ce_out_cache_line <= ce_direction? r_local_data : 0;

	me_out_msg        <= me_direction? r_local_data[(INFO_WIDTH-1) -: MSG_BITS] : 0; 
	me_out_address    <= me_direction? r_local_data[(INFO_WIDTH-MSG_BITS-1) -: ADDRESS_BITS] : 0;
	me_out_cache_line <= me_direction? r_local_data : 0;

end 

/*
 always @ (posedge clock) begin 
	$display ("----------------------------- Network Interface %d-------------------------------", CORE);   
	$display ("M Rp_write [%b]\t| Rp_read [%b]\t| Rq_write [%b]\t| Rq_read [%b]", me_resp_write, 
				me_resp_read, me_req_write, me_req_read); 
	$display ("M Req_valid [%b]\t| Req_data [%h]", me_req_valid, me_req_data); 
	$display ("M Resp_valid [%b]\t| Resp_data [%h]", me_resp_valid, me_resp_data); 
	$display ("C Rp_write [%b]\t| Rp_read [%b]\t| Rq_write [%b]\t| Rq_read [%b]", ce_resp_write, 
				ce_resp_read, ce_req_write, ce_req_read); 
	$display ("C Req_valid [%b]\t| Req_data [%h]",  ce_req_valid, ce_req_data); 
	$display ("C Resp_valid [%b]\t| Resp_data [%h]", ce_resp_valid, ce_resp_data); 	
	$display (""); 
	$display ("ME_IN_MSG [%d]\t| Dest [%d]\t| Address [%h]", me_in_msg, me_in_cache_line[(LINE_BITS -1) -: CACHE_ID_BITS], me_in_address);
	$display ("ME_OUT_MSG [%d]\t| Src [%d]\t| Address [%h]", me_out_msg, me_out_cache_line[(LINE_BITS -1) -: CACHE_ID_BITS], me_out_address);
	$display ("");	
	$display ("CE_IN_MSG [%d]\t| Dest [%d]\t| Address [%h]", ce_in_msg, ce_in_cache_line[(LINE_BITS -1) -: CACHE_ID_BITS], ce_in_address);
	$display ("CE_OUT_MSG [%d]\t| Src [%d]\t| Address [%h]", ce_out_msg, ce_out_cache_line[(LINE_BITS -1) -: CACHE_ID_BITS], ce_out_address);
	$display (""); 
	$display ("S State [%d]\t\t| found [%b]", s_state, found);
	$display ("free_vc[%b]\t\t| free_port  [%b]\t| message [%d]",free_vc, free_port , s_flit_payload[(PAY_WIDTH - 1) -:MSG_BITS]);
	$display ("destination [%d]\t| sub_flow [%b]",s_destination , s_sub_flow );
	$display ("flit_type[%d]\t\t| flit_vc [%b]\t\t| flit_payload [%h]",s_flit_type, s_flit_vc , s_flit_payload);
	$display ("counter[%d]\t\t| s_local_data [%h]",s_counter, s_local_data);
	$display (""); 
	$display ("R State [%d]\t\t| Rec vc[%b]\t\t| Rec port  [%b]\t\t| counter[%d]", r_state, vc, port, r_counter);
	$display ("Source[%d]\t\t| message [%d]\t\t| sub_flow [%b]", r_source, r_flit_pay_msg, r_sub_flow);
	$display ("flit_type[%d]\t\t| flit_vc [%b]\t\t| flit_pay_data [%h]", r_flit_type, r_flit_vc , r_flit_pay_data);
	$display ("direction[%d]\t\t| local_data [%h]", r_direction, r_local_data);
	$display (""); 
	$display ("From Core Flit [%h] ", from_core_flit);
	$display ("From Core Valid [%b] ", v_from_core);
	$display ("From Core Empty [%b]\t| From Core Full [%b] ", from_core_empty, from_core_full);
	$display ("To Core Flit [%h] ", to_core_flit);
	$display ("To Core Valid [%b] ", v_to_core);
	$display ("To Core Empty [%b]\t| To Core Full [%b] ", to_core_empty, to_core_full);
	$display ("-----------------------------------------------------------------------------------");
 end 	
//*/

endmodule 


